/* librist. Copyright © 2019 SipRadius LLC. All right reserved.
 * Author: Daniele Lacamera <root@danielinux.net>
 * Author: Sergio Ammirata, Ph.D. <sergio@ammirata.net>
 *
 * SPDX-License-Identifier: BSD-2-Clause
 */

#ifndef __LIBEVSOCKET
#define __LIBEVSOCKET

#include "common/attributes.h"

struct evsocket_event;
struct evsocket_ctx;

#endif

